/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <sstream>
#include <cmath>
#include "gl-headers.h"
#include "composite-test.h"
#include "options.h"
#include "log.h"
#include "util.h"
#include "render-object.h"
#include "texture.h"
#include "composite-window-ximage.h"

using std::string;
using std::map;
using std::list;
using LibMatrix::vec2;

//
// A RenderObject that blurs the target it is drawn to.
//
class RenderWindowBlur : public RenderObject
{
public:
    RenderWindowBlur(unsigned int passes, unsigned int radius, bool separable) :
        RenderObject(),
        passes_(passes),
        radius_(radius),
        separable_(separable),
        window_contents_() {}

    virtual void init(Program& program)
    {
        RenderObject::init(program);
        window_contents_.init(program);
    }

    virtual void set_background(unsigned int background_texture)
    {
        window_contents_.set_background(background_texture);
    }

    void refresh_window(unsigned int background_texture)
    {
        glClearColor(0.0, 0.0, 0.0, 0.0);
        window_contents_.set_background(background_texture);
        window_contents_.clear();
    }

    virtual void release()
    {
        window_contents_.release();
        RenderObject::release();
    }

    virtual void resize(const vec2& size)
    {
        RenderObject::resize(size);
        window_contents_.resize(size);
    }

    virtual void render_to(RenderObject& target)
    {
        if (separable_) {
            Program& blur_program_h1 = blur_program_h(target.size().x());
            Program& blur_program_v1 = blur_program_v(target.size().y());

            for (unsigned int i = 0; i < passes_; i++) {
                render_from(target, blur_program_h1);
                RenderObject::render_to(target, blur_program_v1);
            }
        }
        else {
            Program& blur_program1 = blur_program(target.size().x(), target.size().y());

            for (unsigned int i = 0; i < passes_; i++) {
                if (i % 2 == 0)
                    render_from(target, blur_program1);
                else
                    RenderObject::render_to(target, blur_program1);
            }

            if (passes_ % 2 == 1)
                RenderObject::render_to(target);
        }

        // Blend the window contents with the target texture.
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        window_contents_.position(position());
        window_contents_.render_to(target);
        glDisable(GL_BLEND);
    }

private:
    enum BlurDirection {
        BlurDirectionHorizontal,
        BlurDirectionVertical,
        BlurDirectionBoth
    };

    void render_from(RenderObject& target, Program& program)
    {
        vec2 final_pos(pos_ + size_);
        vec2 ll_tex(target.normalize_texcoord(pos_));
        vec2 ur_tex(target.normalize_texcoord(final_pos));

        static const GLfloat position_blur[2 * 4] = {
            -1.0, -1.0,
             1.0, -1.0,
            -1.0,  1.0,
             1.0,  1.0,
        };
        GLfloat texcoord_blur[2 * 4] = {
            ll_tex.x(), ll_tex.y(),
            ur_tex.x(), ll_tex.y(),
            ll_tex.x(), ur_tex.y(),
            ur_tex.x(), ur_tex.y(),
        };

        make_current();
        glBindTexture(GL_TEXTURE_2D, target.texture());
        draw_quad_with_program(position_blur, texcoord_blur, program);
    }

    Program& blur_program(unsigned int w, unsigned int h)
    {
        /*
         * If the size of the window has changed we must recreate
         * the shader to contain the correct texture step values.
         */
        if (blur_program_dim_.x() != w || blur_program_dim_.y() != h ||
            !blur_program_.ready())
        {
            blur_program_dim_.x(w);
            blur_program_dim_.y(h);

            blur_program_.release();

            ShaderSource vtx_source;
            ShaderSource frg_source;
            create_blur_shaders(vtx_source, frg_source, radius_,
                                radius_ / 3.0, BlurDirectionBoth);
            frg_source.add_const("TextureStepX", 1.0 / w);
            frg_source.add_const("TextureStepY", 1.0 / h);
            load_shaders_from_strings(blur_program_, vtx_source.str(),
                                      frg_source.str());
        }

        return blur_program_;
    }

    Program& blur_program_h(unsigned int w)
    {
        /*
         * If the size of the window has changed we must recreate
         * the shader to contain the correct texture step values.
         */
        if (blur_program_dim_.x() != w ||
            !blur_program_h_.ready())
        {
            blur_program_dim_.x(w);

            blur_program_h_.release();

            ShaderSource vtx_source;
            ShaderSource frg_source;
            create_blur_shaders(vtx_source, frg_source, radius_,
                                radius_ / 3.0, BlurDirectionHorizontal);
            frg_source.add_const("TextureStepX", 1.0 / w);
            load_shaders_from_strings(blur_program_h_, vtx_source.str(),
                                      frg_source.str());
        }

        return blur_program_h_;
    }

    Program& blur_program_v(unsigned int h)
    {
        /*
         * If the size of the window has changed we must recreate
         * the shader to contain the correct texture step values.
         */
        if (blur_program_dim_.y() != h ||
            !blur_program_v_.ready())
        {
            blur_program_dim_.y(h);

            blur_program_v_.release();

            ShaderSource vtx_source;
            ShaderSource frg_source;
            create_blur_shaders(vtx_source, frg_source, radius_,
                                radius_ / 3.0, BlurDirectionVertical);
            frg_source.add_const("TextureStepY", 1.0 / h);
            load_shaders_from_strings(blur_program_v_, vtx_source.str(),
                                      frg_source.str());
        }

        return blur_program_v_;
    }

    static void
    create_blur_shaders(ShaderSource& vtx_source, ShaderSource& frg_source,
                        unsigned int radius, float sigma, BlurDirection direction);
    LibMatrix::uvec2 blur_program_dim_;
    Program blur_program_;
    Program blur_program_h_;
    Program blur_program_v_;
    unsigned int passes_;
    unsigned int radius_;
    bool separable_;
    RenderClearImage window_contents_;
};

void
RenderWindowBlur::create_blur_shaders(ShaderSource& vtx_source,
    ShaderSource& frg_source, unsigned int radius, float sigma,
    BlurDirection direction)
{
    vtx_source.append_file(GLCOMPBENCH_DATA_PATH"/desktop.vert");
    frg_source.append_file(GLCOMPBENCH_DATA_PATH"/desktop-blur.frag");

    /* Don't let the gaussian curve become too narrow */
    if (sigma < 1.0)
        sigma = 1.0;

    unsigned int side = 2 * radius + 1;

    for (unsigned int i = 0; i < radius + 1; i++) {
        float s2 = 2.0 * sigma * sigma;
        float k = 1.0 / std::sqrt(M_PI * s2) * std::exp( - (static_cast<float>(i) * i) / s2);
        std::stringstream ss_tmp;
        ss_tmp << "Kernel" << i;
        frg_source.add_const(ss_tmp.str(), k);
    }

    std::stringstream ss;
    ss << "result = " << std::endl;

    if (direction == BlurDirectionHorizontal) {
        for (unsigned int i = 0; i < side; i++) {
            int offset = static_cast<int>(i - radius);
            ss << "texture2D(Texture0, TextureCoord + vec2(" <<
                  offset << ".0 * TextureStepX, 0.0)) * Kernel" <<
                  std::abs(offset) << " +" << std::endl;
        }
        ss << "0.0 ;" << std::endl;
    }
    else if (direction == BlurDirectionVertical) {
        for (unsigned int i = 0; i < side; i++) {
            int offset = static_cast<int>(i - radius);
            ss << "texture2D(Texture0, TextureCoord + vec2(0.0, " <<
                  offset << ".0 * TextureStepY)) * Kernel" <<
                  std::abs(offset) << " +" << std::endl;
        }
        ss << "0.0 ;" << std::endl;
    }
    else if (direction == BlurDirectionBoth) {
        for (unsigned int i = 0; i < side; i++) {
            int ioffset = static_cast<int>(i - radius);
            for (unsigned int j = 0; j < side; j++) {
                int joffset = static_cast<int>(j - radius);
                ss << "texture2D(Texture0, TextureCoord + vec2(" <<
                      ioffset << ".0 * TextureStepX, " <<
                      joffset << ".0 * TextureStepY))" <<
                      " * Kernel" << std::abs(ioffset) <<
                      " * Kernel" << std::abs(joffset) << " +" << std::endl;
            }
        }
        ss << " 0.0;" << std::endl;
    }

    frg_source.replace("$CONVOLUTION$", ss.str());
}

//
// Private structure used to avoid contaminating composite-test.h with all of
// the CompositeTestSimpleBlur internal classes.
//
struct BlurPrivate
{
    RenderScreen screen;
    RenderClearImage desktop;
    map<unsigned int, RenderWindowBlur*> windowMap;
    int screen_width;
    int screen_height;
    double lastUpdateTime;
    float window_scale_factor;

    BlurPrivate() : 
        desktop(GLCOMPBENCH_DATA_PATH"/background.png"),
        screen_width(0),
        screen_height(0),
        lastUpdateTime(0),
        window_scale_factor(0.4) {}

    ~BlurPrivate() {}
};

CompositeTestSimpleBlur::CompositeTestSimpleBlur() :
    CompositeTestSimpleBase("blur",
                            GLCOMPBENCH_DATA_PATH"/desktop.vert",
                            GLCOMPBENCH_DATA_PATH"/desktop.frag")
{
    options_["use-vbo"] = CompositeTest::Option("use-vbo", "false",
        "Whether to use VBOs for rendering [true,false]");
    options_["window-size"] = CompositeTest::Option("window-size", "0.5",
        "a scaling factor for the window size (relative to screen screen size) [0.0 - 0.5]");
    options_["passes"] = CompositeTest::Option("passes", "1",
        "the number of effect passes (effect dependent)");
    options_["blur-radius"] = CompositeTest::Option("blur-radius", "5",
        "the blur effect radius (in pixels)");
    options_["separable"] = CompositeTest::Option("separable", "true",
        "use separable convolution for the blur effect");
    priv_ = new BlurPrivate();
}

CompositeTestSimpleBlur::~CompositeTestSimpleBlur()
{
    for (map<unsigned int, RenderWindowBlur*>::iterator winIt = priv_->windowMap.begin();
         winIt != priv_->windowMap.end();
         winIt++)
    {
        RenderWindowBlur* ro = winIt->second;
        delete ro;
    }
    delete priv_;
}

//
// Width/height values used to compute buffer sizes are likely to generate
// errors in the GL implementation if they are not at least 1, so make sure we
// have sane values.
//
static void
clamp_size(vec2& size)
{
    if (size.x() < 1.0)
    {
        size.x(1.0);
    }
    if (size.y() < 1.0)
    {
        size.y(1.0);
    }
}

void
CompositeTestSimpleBlur::prepare_for_run(list<CompositeWindow*>& window_list)
{
    // See how our options tell us to behave...
    priv_->window_scale_factor = Util::fromString<float>(options_["window-size"].value);
    unsigned int passes(Util::fromString<unsigned int>(options_["passes"].value));
    unsigned int radius(Util::fromString<unsigned int>(options_["blur-radius"].value));
    bool separable(options_["separable"].value == "true");

    // Ensure we get a transparent clear color for all following operations
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glDisable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);

    priv_->screen.init(program_);
    priv_->desktop.init(program_);

    unsigned int visible_windows(num_visible_windows(window_list));
    float angular_step(2 * M_PI / visible_windows);
    vec2 screen_size = priv_->screen.size();
    unsigned int i(0);
    for(list<CompositeWindow*>::iterator iter = window_list.begin();
        iter != window_list.end();
        ++iter)
    {
        CompositeWindow* win = *iter;
        if (!win->get_texture().is_valid())
        {
            continue;
        }
        vec2 window_size(win->width(), win->height());
        window_size *= priv_->window_scale_factor;
        clamp_size(window_size);
        vec2 corner_offset(window_size / 2.0);
        RenderWindowBlur* ro = new RenderWindowBlur(passes, radius, separable);
        ro->init(program_);
        ro->set_background(win->get_texture().i);
        vec2 center(screen_size.x() * (0.5 + 0.25 * cos(i * angular_step)),
                    screen_size.y() * (0.5 + 0.25 * sin(i * angular_step)));
        ro->position(center - corner_offset);
        ro->resize(window_size);
        priv_->windowMap.insert(make_pair(win->get_xwindow(), ro));
        i++;
    }
    //
    // Ensure the screen is the current rendering target (it might have changed
    // to a FBO in the previous steps).
    //
    priv_->screen.make_current();
}

void
CompositeTestSimpleBlur::reshape(int width, int height)
{
    CompositeTestSimpleBase::reshape(width, height);
    vec2 screen_size(width, height);
    priv_->screen.resize(screen_size);
    priv_->desktop.resize(screen_size);
}

void
CompositeTestSimpleBlur::draw(list<CompositeWindow *> &window_list)
{
    unsigned int visible_windows(num_visible_windows(window_list));
    float angular_step(2 * M_PI / visible_windows);
    vec2 screen_size = priv_->screen.size();
    unsigned int i(0);

    // Ensure we get a transparent clear color for all following operations.
    glClearColor(0.0, 0.0, 0.0, 0.0);
    priv_->desktop.clear();

    for(list<CompositeWindow*>::iterator cwIt = window_list.begin();
        cwIt != window_list.end();
        ++cwIt)
    {
        CompositeWindow* cw = *cwIt;
        if (!cw->get_texture().is_valid())
        {
            continue;
        }
        unsigned int winHandle(cw->get_xwindow());
        map<unsigned int, RenderWindowBlur*>::iterator roIt = priv_->windowMap.find(winHandle);
        if (roIt == priv_->windowMap.end())
        {
            Log::debug("Failed to find window 0x%x in window map\n", winHandle);
            continue;
        }
        RenderWindowBlur* ro = roIt->second;
        ro->refresh_window(cw->get_texture().i);
        vec2 window_size(cw->width(), cw->height());
        window_size *= priv_->window_scale_factor;
        clamp_size(window_size);
        vec2 corner_offset(window_size / 2.0);
        vec2 center(screen_size.x() * (0.5 + 0.25 * cos(i * angular_step)),
                    screen_size.y() * (0.5 + 0.25 * sin(i * angular_step)));
        ro->position(center - corner_offset);
        ro->resize(window_size);
        ro->render_to(priv_->desktop);
        i++;
    }

    // Present the results to the screen.
    priv_->desktop.render_to(priv_->screen);
}
