/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#ifndef COMPOSITE_TEST_H_
#define COMPOSITE_TEST_H_

#include <string>
#include <list>
#include <map>
#include <pixman.h>
#include <X11/extensions/Xrender.h>

#include "stack.h"
#include "program.h"
#include "composite-window.h"
#include "vbo.h"
#include "shader-source.h"

class CompositeCanvas;

class CompositeTest
{
public:
    CompositeTest(const std::string& test_name, const std::string &type);

    virtual void init() {}
    virtual void deinit() {}
    virtual void prepare_for_run(std::list<CompositeWindow*> &window_list)
    {
        if (window_list.empty())
            return;
    }
    virtual void draw(std::list<CompositeWindow*> &window_list)
    {
        (void)window_list;
    }
    virtual void cleanup() {}
    virtual void reshape(int width, int height)
    {
        (void)width;
        (void)height;
    }
    const std::string& name() const { return name_; }

    struct Option {
        Option(const std::string &name, const std::string &val, const std::string &desc) :
            name(name), value(val), default_value(val), description(desc) {}
        Option() {}
        std::string name;
        std::string value;
        std::string default_value;
        std::string description;
    };

    virtual bool set_option(const std::string &opt, const std::string &val);
    void reset_options();
    bool set_option_default(const std::string &opt, const std::string &val);
    const std::map<std::string, Option> &options() { return options_; }

    const std::string &type() { return type_; }
    void set_canvas(CompositeCanvas *canvas) { canvas_ = canvas; }
    CompositeCanvas *canvas() { return canvas_; }

    static CompositeTest &dummy()
    {
        static CompositeTest dummy_test("", "");
        return dummy_test;
    }

protected:
    CompositeTest();
    //
    // Gets the number of windows that are visible.
    //
    // @param window_list the window list to search in
    //
    // @return the number of visible windows
    //
    int num_visible_windows(std::list<CompositeWindow *> &window_list);
    std::string name_;
    std::map<std::string, Option> options_;
    std::string type_;
    bool initialized_;
    CompositeCanvas *canvas_;
};

class CompositeTestDefaultOptions : public CompositeTest
{
public:
    CompositeTestDefaultOptions() : CompositeTest("", "") {}

    virtual void prepare_for_run(std::list<CompositeWindow*> &window_list);
    virtual bool set_option(const std::string &opt, const std::string &val);

protected:
    std::list<std::pair<std::string, std::string> > default_options_;
};

class CompositeTestSimpleBase : public CompositeTest
{
public:
    CompositeTestSimpleBase(const std::string& test_name,
                            const std::string& vs_filename,
                            const std::string& fs_filename);

    virtual void init();
    virtual void deinit();
    virtual void prepare_for_run(std::list<CompositeWindow*> &window_list);
    virtual void cleanup();
    virtual void reshape(int width, int height);

protected:
    Program program_;
    std::string vs_filename_;
    std::string fs_filename_;
    std::string vertex_shader_;
    std::string fragment_shader_;
    int vertexIndex_;
    int texcoordIndex_;
    VertexBufferData vboData_;
    static const std::string model_view_matrix_name_;
    static const std::string projection_matrix_name_;
    static const std::string texture0_name_;
    static const std::string texcoord_name_;
    static const std::string position_name_;

    LibMatrix::Stack4 projection_matrix;
    LibMatrix::Stack4 model_view_matrix;
    virtual bool init_shaders(ShaderSource& vtx, ShaderSource& frg);
};

class CompositeTestSimpleDefault : public CompositeTestSimpleBase
{
public:
    CompositeTestSimpleDefault() : 
        CompositeTestSimpleBase("default",
                                GLCOMPBENCH_DATA_PATH"/default.vert",
                                GLCOMPBENCH_DATA_PATH"/default.frag")
        {}

    virtual void draw(std::list<CompositeWindow*> &window_list);
    virtual bool init_shaders(ShaderSource& vtx, ShaderSource& frg);
};

class CompositeTestSimpleBrick : public CompositeTestSimpleBase
{
public:
    CompositeTestSimpleBrick() :
        CompositeTestSimpleBase("brick",
                                GLCOMPBENCH_DATA_PATH"/brick.vert",
                                GLCOMPBENCH_DATA_PATH"/brick.frag")
        {}

    virtual void draw(std::list<CompositeWindow*> &window_list);
    virtual void prepare_for_run(std::list<CompositeWindow*> &window_list);
private:
    LibMatrix::vec4 lightPos_;
    static const LibMatrix::vec3 mortarColor_;
    static const LibMatrix::vec2 brickSize_;
    static const LibMatrix::vec2 brickPct_;
    static const std::string brickSizeName_;
    static const std::string brickPctName_;
    static const std::string mortarColorName_;
    static const std::string lightPosName_;
};

class CompositeTestPixman : public CompositeTest
{
public:
    CompositeTestPixman();

    virtual void init();
    virtual void deinit();
    virtual void prepare_for_run(std::list<CompositeWindow*> &window_list);
    virtual void draw(std::list<CompositeWindow*> &window_list);

private:
    pixman_image_t *half_transparent_mask_;
    pixman_filter_t pixman_filter_;
};

class CompositeTestXRender : public CompositeTest
{
public:
    CompositeTestXRender();

    virtual void init();
    virtual void deinit();
    virtual void prepare_for_run(std::list<CompositeWindow*> &window_list);
    virtual void draw(std::list<CompositeWindow*> &window_list);

private:
    Picture half_transparent_mask_;
    const char *xrender_filter_;
    
};

class CompositeTestSimpleFade : public CompositeTestSimpleBase
{
    static const std::string fade_bias_name_;
public:
    CompositeTestSimpleFade() : 
        CompositeTestSimpleBase("fade",
                                GLCOMPBENCH_DATA_PATH"/default.vert",
                                GLCOMPBENCH_DATA_PATH"/default.frag")
        {}

    virtual void draw(std::list<CompositeWindow*> &window_list);
    virtual bool init_shaders(ShaderSource& vtx, ShaderSource& frg);
};

struct BlurPrivate;

class CompositeTestSimpleBlur : public CompositeTestSimpleBase
{
    static const std::string fade_bias_name_;
    BlurPrivate* priv_;
public:
    CompositeTestSimpleBlur();
    ~CompositeTestSimpleBlur();

    virtual void draw(std::list<CompositeWindow*> &window_list);
    virtual void reshape(int width, int height);
    virtual void prepare_for_run(std::list<CompositeWindow*> &window_list);
};

#endif // COMPOSITE_TEST_H_
